# Copyright (C) 2010 Alejandro Ochoa, Singh Research Group, Princeton University
# This file is part of dPUC.
# dPUC is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# dPUC is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with dPUC.  If not, see <http://www.gnu.org/licenses/>.

package General;
use POSIX ();
use PerlIO::gzip; # my entire Pfam database is gzipped
use strict;

# constants
my $log2 = log(2);

sub currDate {
    return POSIX::strftime "%F %T %Z", localtime;
}

sub randomIdGen {
    # this sub generates a random ID of a desired length, using only alphanumerical characters (so they play nice with HTML, URIs, etc)
    my ($length) = @_; # this is the desired length
    $length = 10 unless defined $length; # default length
    my @letters = ('a'..'z', 'A'..'Z', 0 .. 9); # things we can use without messing up URIs
    # start new random ID
    my $id = '';
    foreach (1..$length) {
	$id .= $letters[rand @letters]; # this chooses a random letter each time
    }
    return $id; # done!
}

sub randomIdGenUnique {
    # this sub generates a random ID with randomIdGen(), then compares it to a list of existing IDs and iterates until a new one is finally produces
    # see randomIdGen() for more details
    my ($length, $ids) = @_; # this is the desired length, and the list of existing IDs
    # make an initial ID
    my $id = randomIdGen($length);
    # start iteration untile we have a unique ID
    until (General::binarySearch($id, $ids) == -1) {
	$id = randomIdGen($length);
    }
    # done, return
    return $id;
}

# operations on sets/lists

sub add {
    # for strings only, case sensitive
    # modifies (updates) array referenced in $original but not $toAdd
    # Note this is not the same as a binary union, since it does not check for repetitions in the $original array.
    # This is overcome in union() since the starting array is empty.
    # returns number of times something was already in the list
    my ($original, $toAdd) = @_;
    my $c = 0; # counts number of times something was already in the list
    foreach my $toAddElem (@$toAdd) {
	next if $toAddElem eq '';
	if (isIn($toAddElem, $original)) { $c++; }
	else { push @$original, $toAddElem; }
    }
    return $c;
}

sub isIn {
    # for strings only
    # does not modify input, returns boolean
    my ($str, $list) = @_;
    foreach my $i (@$list){ return 1 if $str eq $i; }
    return 0;
}

sub binarySearch {
    # search array of strings @$a for $x
    # return index where found or -1 if not found
    # IMPORTANT: assumes list @$a is already sorted ascending, it doesn't work otherwise!!!
    # we don't sort here cause it would run the sort at every single call, but we can probably do it much less often than that...
    # based on code stolen from
    # http://staff.washington.edu/jon/dsa-perl/bsearch
    # I tried a further optimization based on wikipedia article (defer equality checking till end), but that made it much slower in my case
    my ($x, $a) = @_;
    my ($l, $u) = (0, @$a - 1); # lower, upper end of search interval
    my $i; # index of probe
    my $s; # string at probe (initialization factored out of tight loop)
    while ($l <= $u) {
	$i = int(($l + $u)/2); # choose the index right at the middle of our range
	$s = $a->[$i]; # get the string here
	if ($s lt $x) { $l = $i+1; } # update bounds accordingly (next line too)
	elsif ($s gt $x) { $u = $i-1; }
	else { return $i; } # found
    }
    return -1; # not found
}

sub isIntersectionBinarySearch {
    # for strings
    # this is similar to the above with two differences
    # 1) only one of the lists has to be sorted (the second one)
    # 2) returns a boolean, so stops looking after we've found one item that intersects (faster???)
    my ($l1, $l2) = @_; # lists.
    foreach my $i (@$l1) {
	return 1 if binarySearch($i, $l2) != -1;
    }
    return 0;
}

sub paraSortNum {
    # this sorts multiple arrays in parallel by sorting the first one numerically ascending
    # adapted from http://www.perlmonks.org/?node_id=350180
    # NOTE: I could probably also try a schwartzian version, which should be easy to write for two arrays (not for n) and it might be faster than this, who knows
    # http://en.wikipedia.org/wiki/Schwartzian_transform
    my ($arrayMaster, @arrayRefs) = @_;
    # make the list of indeces
    my @indexes = (0 .. $#$arrayMaster);
    # this sorts the indexes so that the master array is sorted numerically ascending
    @indexes = sort { $arrayMaster->[$a] <=> $arrayMaster->[$b] } @indexes;
    # finally, this takes care of sorting all the arrays appropriately before returning
    return map { [ @$_[ @indexes ] ] } ( $arrayMaster, @arrayRefs );
}

# some simple math

sub log2 {
    # uses the stored value of log10 to make repeated calculations very fast
    return log(shift)/$log2;
}

sub t {
    # maps $i, $j into $t in a triangle matrix of any size!
    # maps contiguously if we do: i in 0..total, j in 0..i (allowing diagonal), good for loops so we don't have to calculate this, but we do need to calculate it to map back
    # in current form, assumes i>=j
    my ($i, $j) = @_;
    return $j + $i*($i+1)/2;
}

# I/O, originally from the KEATING LAB

sub getInFh {
    my $f = shift;
    die "Undefined or empty file name passed!" if !defined $f || $f eq '';
    my $fh;
    if ($f eq 'STDIN' || $f eq '-') { $fh = \*STDIN; } # useful hack
    elsif (-r $f) {
	if ($f =~ /\.gz$/i) {
	    open($fh, '<:gzip', $f) || die "Could not open for reading $f: $!";
	}
	else {
	    open($fh, $f) || die "Could not open for reading $f: $!";
	}
    }
    else {
	if (-r "$f.bz2") {
	    open($fh, "bzip2 -dc $f.bz2 |") || die "Could not open for reading $f.bz2: $!";
	}
	elsif (-r "$f.gz") {
	    open($fh, '<:gzip', "$f.gz") || die "Could not open for reading $f.gz: $!";
	}
	else { die "Error: file '$f' doesn't exist (?). Could not open for reading."; }
    }
    return $fh;
}

sub getOutFh {
    my ($f, $comp) = @_;
    die "Undefined or empty file name passed!" if !defined $f || $f eq '';
    $comp = '' unless defined $comp;
    my $fh;
    if ($f eq 'STDOUT' || $f eq '-') { $fh = \*STDOUT; } # useful hack.  Note STDOUT is never gzipped, even if that option is passed!
    elsif ($f =~ /\.gz$/i) {
	open($fh, '>:gzip', $f) || die "Could not open for writing $f: $!";
    }
    elsif ($comp =~ /bzip2/i) {
	open($fh, "| bzip2 -zcf > $f.bz2") || die "Could not open for writing $f.bz2: $!";
    }
    elsif ($comp =~ /gzip/i) {
	open($fh, '>:gzip', "$f.gz") || die "Could not open for writing $f.gz: $!";
    }
    else {
	open($fh, '>', $f) || die "Could not open for writing $f: $!";
    }
    return $fh;
}

sub getAppFh {
    # this shouldn't work with compression, so we just open it normal
    my ($f) = @_;
    die "Undefined or empty file name passed!" if !defined $f || $f eq '';
    open(my $fh, '>>', $f) || die "Could not open for appending $f: $!";
    return $fh;
}

sub getFilesInDir {
    # returns the directory's non-hidden, non-trivial files, sorted
    # if $flag eq 'dirs', returns directories, else (including 'files' and undef) returns files
    my ($dir, $flag) = @_;
    opendir(my $dh, $dir) || die "Error: Could not open dir $dir: $!";
    my @things = readdir $dh;
    closedir $dh;
    my @files = ();
    $dir .= '/' unless $dir =~ /\/$/ || $dir eq ''; # add slash unless it is already there or $dir is empty
    foreach my $t (@things) {
	if (defined $flag && $flag eq 'dirs') {
	    push @files, $t if -d $dir.$t && !($t =~ /^\./);
	}
	else { # if (!$flag || $flag eq 'files'), but with more
	    push @files, $t if -f $dir.$t && !($t =~ /^\./);
	}
    }
    return sort @files;
}

# other IO wrappers that handle compression transparently 

sub fileExists {
    # for a single file, returns a boolean that tells us if it, or a compressed version, exist on the disk
    my $f = shift;
    if (-f $f) { return 1; }
    elsif (-f $f.'.gz') { return 1; }
    else { return 0; }
}

sub myRename {
    # move fi to fo, or their compressed versions
    my ($fi, $fo) = @_;
    if (-f $fi) { rename($fi, $fo) || die "Error: could not rename file: $!"; } # if $fi is regular, then move it and $fo will be as required
    elsif (-f $fi.'.gz') { rename($fi.'.gz', $fo.'.gz') || die "Error: could not rename file: $!"; } # if $fi is compressed, move it to $fo and keep it compressed.  This assumes the extensions were missing on both sides
}

sub readHashFile {
    # very retarded, expects a file with two columns, one is the keys, the other the values, tab delimited.  Returns the desired hash's reference.  Ignores problems with repeated keys, overwrites so that only the last instance counts.
    # unlike the printing function, this one returns all values as text, even if they used to be arrays textified with "join semicolons".
    my $fi = shift;
    my %hash;
    my $fhi = getInFh($fi);
    while (<$fhi>) {
	chomp;
	my ($key, $value) = split /\t/;
	$hash{$key} = $value;
    }
    close $fhi;
    return \%hash;
}

sub printHashFile {
    # prints a tab-delimited file with two columns: keys and values.  Input is a file name and the hash to print.
    # $comp (to use compression, e.g. 'gzip') is optional, default is no compression
    # $header (for R to read as table, for values only) is also optional, default is no header
    my ($fo, $hash, $comp, $header) = @_;
    my $fho = getOutFh($fo, $comp);
    print $fho "\t$header\n" if $header;
    while (my ($key, $value) = each %$hash) {
	# dumb check, sometimes value is an array, always join with semicolons in that case
	$value = join ';', @$value if ref $value eq 'ARRAY';
	print $fho "$key\t$value\n";
    }
    close $fho;
}

1;
