# Copyright (C) 2010 Alejandro Ochoa, Singh Research Group, Princeton University
# This file is part of dPUC.
# dPUC is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# dPUC is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with dPUC.  If not, see <http://www.gnu.org/licenses/>.

package Nets;
use lib '.';
use General;
use strict;

# a package that handles simple networks
# and a few extra special subs

sub printNet {
    # compression is an optional parameter
    my ($fo, $net, $comp) = @_;
    my $fho = General::getOutFh($fo, $comp); # these files are big, so I'll compress them.
    while (my ($acc1, $acc2s) = each %$net) {
	while (my ($acc2, $value) = each %$acc2s) {
	    print $fho join("\t", $acc1, $acc2, $value)."\n";
	}
    }
    close $fho;
}

sub loadNet {
    my ($fi, $cut) = @_;
    # loads a net from file
    # assumes net is in our format of $acc1 before $acc2 in dictionary
    # cut means don't load interactions below that value, a filter
    my $isCut = defined $cut; # default is no cut
    my %net;
    my $fhi = General::getInFh($fi);
    while (<$fhi>) {
	chomp;
	my ($acc1, $acc2, $value) = split /\t/;
	$net{$acc1}{$acc2} = $value if !$isCut || $value > $cut; # add only if cutoff is passed
    }
    close $fhi;
    return \%net;
}

sub scaleNetAndInt {
    # this takes a net with floating point values and scales them by the provided $scoreScale, then turns into integers via rounding
    # net is modified in place by reference
    my ($net, $scoreScale) = @_;
    while (my ($acc1, $acc2s) = each %$net) {
	foreach my $acc2 (keys %$acc2s) {
	    my $s = $acc2s->{$acc2};
	    $acc2s->{$acc2} = int $scoreScale*$s+1/2;
	}
    }
}

sub mapPfamNetAccs2ints {
    # this sub takes a standard net with two pfam accessions mapped to a score, and turns it into a parallel array, one with the accessions turned into integers and combined into a triangle-matrix flat number t(i,j), and another array with the corresponding score
    # this format is optimal for fast queries in C code, but it's not very perlish at all
    # in principle this can be done for SMART and Supfam, but the current code only handles Pfam because of the PF substitution (I guess I could be a little smarter aobut it but meh)
    my ($net) = @_;
    # these are the parallel arrays we want
    my @ts;
    my @tscores;
    while (my ($acc1, $acc2s) = each %$net) {
	my $accInt1 = $acc1;
	$accInt1 =~ s/^PF//;
	while (my ($acc2, $s) = each %$acc2s) {
	    my $accInt2 = $acc2;
	    $accInt2 =~ s/^PF//;
	    my ($i, $j) = (int $accInt1, int $accInt2); # copy ints, but these copies will be sorted
	    ($i, $j) = ($j, $i) if $i < $j; # this forces i to be larger or equal than j, as t(i,j) requires
	    my $t = General::t($i, $j); # flatten with t()
	    push @ts, $t; # add flattened index of Pfam accessions pair
	    push @tscores, $s; # add score in parallel
	}
    }
    # unfortunately, the above procedure doesn't produce a list of @ts that is sorted ascending (important for binary searches!), so we have to sort now, then return
    return General::paraSortNum(\@ts, \@tscores);
}

sub makeAllEdgesBothWays {
    # normally we keep the undirected network so that edges satisfy $acc1 lt $acc2, but for searching purposes we want both directions
    # this makes a new graph, so the old graph is not modified
    my $net = shift;
    my %net; # net net with edges both ways
    # navigate input network
    while (my ($acc1, $acc2s) = each %$net) {
	while (my ($acc2, $value) = each %$acc2s) {
	    # transfer values to new network
	    $net{$acc1}{$acc2} = $value;
	    $net{$acc2}{$acc1} = $value;
	}
    }
    return \%net; # return new net
}

1;
