#!/usr/bin/perl
# Copyright (C) 2010 Alejandro Ochoa, Singh Research Group, Princeton University
# This file is part of dPUC.
# dPUC is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# dPUC is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with dPUC.  If not, see <http://www.gnu.org/licenses/>.

use YAML (); # nice for loading/saving data structures
use Mail::Sendmail ();
use Proc::PID::File;
use Proc::Daemon;
use Text::Unaccent::PurePerl ();
use lib '.';
use General;
use FilePaths;
use ParseAln;
use ContextDaemon;
use lib '/home/viiia/sitius/scripts/';
use Dico;
use strict;
use warnings;
use utf8;

# file reading is working, as well as writing, but not moving!  Email worked!

# this script loads all necessary data for processing, then waits in the background for new jobs to appear on a special directory, and runs them as they come.  No defaults are applied here, all data has to be present.

# params
my $dpucDir = '/var/dPUC/'; # main dPUC directory, with logs and runs
my $queueDir = $dpucDir.'queue/'; # queue of jobs that have been submitted
my $predsDir = $dpucDir.'preds/'; # dir with final predictions for users to see online
my $logsDir = $dpucDir.'logs/'; # logs dir
my $logFile = $logsDir.logFilePath(); # file has a name per month started, but might contain more than a month's contents if my server is running uninterrupted for several months
my $fileCheckTime = 10; # when not running HMMER, check for new files at this interval, in seconds
my $idLength = 10; # length in characters of random ID we want for each session

# Since HMMER2 output might be harder to parse for errors, let's check that input makes sense (at least that files exist!)
# ContextDaemon.pm loads a bunch of files right away, before daemonization, so if they're not found my program will die before daemonization!
# these are the only extra things that I have to check for separately... just die if things went wrong
# files that should exist...
die "Error: $FilePaths::pfamLs does not exist!" unless General::fileExists($FilePaths::pfamLs);
die "Error: $FilePaths::pfamFs does not exist!" unless General::fileExists($FilePaths::pfamFs);
# binaries...
checkBinary($FilePaths::hmmpfam);
checkBinary($FilePaths::lpSolve);

# Fork and detach from the parent process
Proc::Daemon::Init;
die "Already running!" if Proc::PID::File->running();

# create main directories unless they're already there
mkdirs($dpucDir, $logsDir, $queueDir, $predsDir);
myConWeb($predsDir); # set regular apache permissions on the directory that we need to see publicly
# start log
my $fhoLog = General::getAppFh($logFile);
msg("Starting dPUC");

# Setup signal handlers so that we have time to cleanup before shutting down
my $keep_going = 1;
$SIG{HUP} = sub { msg("Caught SIGHUP: exiting gracefully"); $keep_going = 0; };
$SIG{INT} = sub { msg("Caught SIGINT: exiting gracefully"); $keep_going = 0; };
$SIG{QUIT} = sub { msg("Caught SIGQUIT: exiting gracefully"); $keep_going = 0; };
$SIG{TERM} = sub { msg("Caught SIGTERM: exiting gracefully"); $keep_going = 0; };

# start an "infinite" loop of file surveilance
while ($keep_going) {
    # check if there's a file to run
    # this sub returns a list, but we're interested in getting the first file only (sorted ascending as string)
    my @files = General::getFilesInDir($queueDir); # get all files first
    @files = grep /\.yml(?:\.gz)?$/, @files; # only pay attention to files with expected extension
    my $fi = $files[0]; # now get the first one
    
    # this means we had no file
    unless (defined $fi) {
	sleep $fileCheckTime; # delay time to check again
	next; # repeat main loop
    }
    msg("Found file $fi");
    
    # input file should be YAML
    my $data = readYaml($queueDir.$fi);
    last unless defined $data; # end program if bad data was passed?
    # get session data
    my $l = $data->{l};
    my $css = $data->{css};
    my $name = $data->{name};
    my $email = $data->{email};
    my $Ecut = $data->{Evalue};
    my $ids = $data->{ids};
    my $id2seq = $data->{id2seq};
    msg("  Data was read from $name ($email), $l, $css, E <= $Ecut, ".(scalar @$ids)." proteins");
    
    # prepare raw output
    my $fo = $fi; $fo =~ s/\.yml(?:\.gz)?$/.out.txt/; # raw output is same as input but with different extension
    my $fho = General::getOutFh($queueDir.$fo, $FilePaths::comp);
    # also nice YAML structure with domains only (no stats and other stuff)
    my %prot2db2data;
    
    # start running each sequence in file, in original order
    foreach my $id (@$ids) {
	# get sequence of this ID
	my $seq = $id2seq->{$id};
	msg("  Processing protein $id"); # : $seq
	
	# do all processing by calling this one function
	# output is an array of very different things, but we don't need to name them all, they're in the same order as the printing function wants them.
	my ($hitsDpuc, $hitsGa, @output) = ContextDaemon::doSeqAll($id, $seq, $Ecut); # , $fhoLog # for debugging only
	# print predictions and stats to output
	ContextDaemon::printOutputText($fho, $id, $Ecut, $hitsDpuc, $hitsGa, @output);
	# put into a YAML structure that other code can process more easily...
	$prot2db2data{$id} = {
	    'Length' => length $seq,
	    'Pfam23' => $hitsGa,
	    'Pfam23dpuc' => $hitsDpuc,
	};
    }
    # output is definitely a file, have to close...
    close $fho;
    
    ### run is now completely done...
    msg("  Protein processing complete");
    
    # first make a unique session ID
    # get existing IDs in predictions directory, they come out sorted already!
    my @sIds = General::getFilesInDir($predsDir);
    # this produces a unique ID
    my $sId = General::randomIdGenUnique($idLength, \@sIds);
    
    # get better names than random ones, this time the whole session is encoded by the directory name so names can be more generic
    my $fiClean = $fi; $fiClean =~ s/^.*\.yml(\.gz)?$/input.yml$1/;
    my $foClean = $fo; $foClean =~ s/^.*\.out\.txt(\.gz)?$/output.out.txt$1/;
    # we should move files to their destination (viewable from the web)
    my $predsIdDir = $predsDir.$sId.'/'; # output directory
    mkdirs($predsIdDir); # make directory unless it was already there
    General::myRename($queueDir.$fi, $predsIdDir.$fiClean); # actually move input and output
    General::myRename($queueDir.$fo, $predsIdDir.$foClean);
    myConWeb($predsIdDir); # set regular apache permissions on the directory that we need to see publicly.  these files were made on a directory that didn't have the right permissions, so we need to set them again
    msg("  Moved data to output dir $sId");
    
    # we should also make the pretty pictures HTML...
    # actually make YAML file that gets transformed by my website's code
    my $dateYaml = General::currDate();
    my %data = (
	'format' => 'proteinDomains',
	'head' => {
	    'title' => {
		'en-us' => 'dPUC - Your Predictions',
		'es-mx' => 'dPUC - Sus predicciones',
		'es-an' => 'dPUC - Las sus predictiones',
		'fr' => 'dPUC - Vos prédictions',
	    },
	    'subtitle' => 'Domain Prediction Using Context',
	    'comments' => {
	        'en-us' => 'Extend your Pfam predictions without loss of precision using domain context!',
		'es-an' => '¡Extienda sus predictiones de Pfam sin perder preçision usando el contexto de dominios!',
		'es-mx' => '¡Extienda sus predicciones de Pfam sin perder precisión usando el contexto de dominios!',
		'fr' => 'Étendez vos prédictions Pfam sans perdre précision employant le contexte de protéine!',
	    },
	    'date' => $dateYaml,
	    'author' => [qw(Alex ManuelLlinás Mona)],
	    'langs' => [qw(en-us es-mx es-an fr)],
	    'thumbnail' => '/dPUC/puck.thumb.png',
	},
	'history' => [
	    {
		'date' => $dateYaml,
		'note' => 'Created file',
	    },
	],
	'data' => {
	    'Evalue' => $Ecut,
	    'proteins' => $ids, # proteins in order
	    'dbs' => ['Length', 'Pfam23', 'Pfam23dpuc'], # types, in order
	    'protein2db2data' => \%prot2db2data,
	}
	);
    printYaml($predsIdDir.'index.yml', undef, \%data); # not compressed
    
    # make params string, hacky but I'd rather keep this separate from Sitivs.pm
    my @params;
    push @params, "l=$l" unless $l eq 'es-mx';
    push @params, "css=$css" unless $css eq 'black';
    my $params = join ';', @params;
    $params = '?'.$params if $params; # no question mark if params was empty
    
    # we should also sent the email to the user notifying job completeness
    # all fields were validated already, just use them...
    my %mail = (
	From => 'dPUC Server <ochoa@princeton.edu>',
	To => "$name <$email>",
	Bcc => 'ochoa@princeton.edu', # send myself a copy, at least for tests?
	Subject => Text::Unaccent::PurePerl::unac_string($Dico::d{'Su petición dPUC está completada!'}{$l}),
#	'Content-Type' => 'text/plain; charset=utf-8', # with our current settings, this prints better as windows encoding?
	Message => "dPUC: Domain Prediction Using Context

$Dico::d{'Aquí están sus resultados'}{$l}:
http://viiia.org/dPUC/preds/$sId/$params

Your results will be available for 60 days.
You may reply to this email if you have any questions.
Citaton: ...
",
	);
    Mail::Sendmail::sendmail(%mail); # could complain, but what's the point???
    msg("  Sent email to user.  Run is complete.");
    
    last; # just for a test
}

# Mark a clean exit in the log
msg("Stopping dPUC");
close $fhoLog;


### SUBS

sub logFilePath {
    # it should be YYYY-MM.txt, this code gets it...
    my $date = General::currDate(); # full date
    ($date) = $date =~ /^(\d{4}-\d{2})/; # only year and month pass
    return $date.'.txt';
}

sub msg {
    # messages always have timestamp
    print $fhoLog General::currDate().": ".(shift)."\n";
}

sub mkdirs {
    # makes a list of directories in order (presumably user fed them so lower directories get made first)
    # NOTE: all paths should be absolute!
    
    # make each directory in input
    foreach my $dir (@_) {
	mkdir $dir unless -d $dir;
    }
}

sub myConWeb {
    # set regular apache permissions on the directory that we need to see publicly
    my $dir = shift;
    system "chcon -R -t httpd_sys_script_ro_t $dir";
}

sub checkBinary {
    my $name = shift;
    # name could be a file, just ask if it's executable
    if (-f $name) {
	die "Error: $name is a file but it is not executable" unless -x $name;
    }
    # maybe we're supposed to get it from the path
    # this uses the linux shell, hacky but whatever
    else {
	# this will give us the real path
	$name = `which $name`;
	chomp $name;
	# test as before
	if (-f $name) {
	    die "Error: $name is a file but it is not executable" unless -x $name;
	}
	# all failed, executable not found
	else { die "Error: $name is not a file nor was it in the path?" }
    }
}

### YAML wrappers (to read/write gzip-compressed YAML files transparently)
# these functions aren't memory smart, could load/dump things bit by bit and be a lot more efficient, but other functions will do that

sub readYaml {
    # if parsing fails, returns undefined and logs error...
    my $fi = shift;
    my $fhi = General::getInFh($fi); # this takes care of compression transparently
    my $lines = '';
    while (<$fhi>) { $lines .= $_; } # concatenate entire file into a string
    # now that text stream is uncompressed, use regular YAML parser to load data
    my $data;
    eval { $data = YAML::Load($lines) };
    msg("Yaml parsing of $fi failed: $@\n") if $@;
    return $data;
}

sub printYaml {
    my $fo = shift;
    my $comp = shift;
    # the rest of the data is in @_, didn't want to waste time copying it in
    # this turns the data into text
    my $lines = YAML::Dump(@_);
    # now print the data, using compression if asked to
    my $fho = General::getOutFh($fo, $comp);
    print $fho $lines;
    close $fho;
}

